package org.processmining.plugins.cheetah.statistics;

import java.util.*;

import org.deckfour.xes.model.*;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class FlowNetwork {
	public HashMap<String, Node> nodes = new HashMap<String, Node>();
	public HashMap<Node, ArrayList<CapacityEdge>> network = new HashMap<Node,ArrayList<CapacityEdge>>();
    public HashMap<CapacityEdge, Integer> flow = new HashMap<CapacityEdge, Integer>();
    public ArrayList<Node> gateways = new ArrayList<Node>();
    public Node start_event = null;
    public Node end_event = null;
	
	public FlowNetwork(XTrace trace) {
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event).toUpperCase();
			if (!event.getAttributes().containsKey("id"))
				continue;
			String id = event.getAttributes().get("id").toString();
			if (activity.equals("CREATE_ACTIVITY"))
				addNode(makeNewNode(id,"NODE"));
			else if (activity.equals("CREATE_START_EVENT"))
				addStartEvent(makeNewNode(id,"NODE"));
			else if (activity.equals("CREATE_END_EVENT"))
				addEndEvent(makeNewNode(id,"NODE"));
			else if (activity.equals("CREATE_XOR"))
				addGateway(makeNewNode(id,"XOR"));
			else if (activity.equals("CREATE_AND"))
				addGateway(makeNewNode(id,"AND"));
			else if (activity.equals("CREATE_EDGE")) {
				String source = event.getAttributes().get("source_node_id").toString();
				String target = event.getAttributes().get("target_node_id").toString();
				addEdge(id, findNode(source), findNode(target));
			}
			else if (activity.equals("DELETE_ACTIVITY"))
				removeNode(findNode(id));
			else if (activity.equals("DELETE_START_EVENT"))
				removeStartEvent(findNode(id));
			else if (activity.equals("DELETE_END_EVENT"))
				removeEndEvent(findNode(id));
			else if (activity.equals("DELETE_XOR") || activity.equals("DELETE_AND"))
				removeGateway(findNode(id));
			else if (activity.equals("DELETE_EDGE")) {
				removeEdge(id);
			}
			else if (activity.equals("RECONNECT_EDGE")) {
				try {
					String source = event.getAttributes().get("source_node_id").toString();
					String target = event.getAttributes().get("target_node_id").toString();
					removeEdge(id, findNode(source));
					addEdge(id, findNode(source), findNode(target));
				}
				catch(Exception ex) { //ignore			
				}
			}
		}
    }
	
	private Node findNode(String id) {
		return nodes.get(id);
	}
	private Node makeNewNode(String id, String type) {
		Node n = new Node(id, type);
		nodes.put(id, n);
		return  n;
	}

	public void addNode(Node node) {
		network.put(node, new ArrayList<CapacityEdge>());
	}
	public void removeNode(Node node) {
        //adj.remove(vertex);
		network.put(node, new ArrayList<CapacityEdge>()); // replace current with empty arraylist
	}

	public void addStartEvent(Node start) {
		addNode(start);
		start_event = start;
	}
	public void removeStartEvent(Node start) {
		removeNode(start);
        start_event = null;
	}
	public void addEndEvent(Node end) {
		addNode(end);
		end_event = end;
	}
	public void removeEndEvent(Node end) {
		removeNode(end);
        end_event = null;
	}
	
	public void addGateway(Node gw) {
		addNode(gw);
		gateways.add(gw);
	}
	public void removeGateway(Node gw) {
        removeNode(gw);
        gateways.remove(gw);
	}
 
    public ArrayList<CapacityEdge> getEdges(Node node) {
        return network.get(node);
    }
 
    public void addEdge(String id, Node source, Node target) {
        if (source.equals(target))
            System.out.println("ERROR");
        CapacityEdge edge = new CapacityEdge(id,source,target,1);
        CapacityEdge redge = new CapacityEdge(id,target,source,0);
        edge.setRedge(redge);
        redge.setRedge(edge);
        network.get(source).add(edge);
        network.get(target).add(redge);
        flow.put(edge, new Integer(0));
        flow.put(redge, new Integer(0));
    }
    public void removeEdge(String id) {
    	for (Node key : network.keySet())
    		for (CapacityEdge edge : network.get(key))
    			if (edge.getId().equals(id))
        			{ removeEdge(edge); return; }
    }
    public void removeEdge(String id, Node source) {
    	ArrayList<CapacityEdge> edges = network.get(source);
    	for (CapacityEdge edge : edges)
    		if (edge.getId().equals(id))
    			{ removeEdge(edge); return; }
    }
    public void removeEdge(CapacityEdge edge) {
    	network.get(edge.getSource()).remove(edge);
    	network.get(edge.getTarget()).remove(edge.getRedge());
		flow.remove(edge);
		flow.remove(edge.getRedge());
    }
    private void resetFlow() {
    	for (CapacityEdge edge : flow.keySet())
    		flow.put(edge, new Integer(0));
    }
 
    public HashMap<CapacityEdge, Integer> findPath(Node source, Node target, HashMap<CapacityEdge, Integer> path) {
        if (source.equals(target))
            return path;
        for (CapacityEdge edge : getEdges(source)) {
            int residual = edge.getCapacity() - flow.get(edge).intValue();
            if (residual > 0 && !path.containsKey(edge)) {
            	path.put(edge, residual);
            	HashMap<CapacityEdge, Integer> result = findPath(edge.getTarget(), target, path); 
                if (!result.isEmpty())
                    return result;
                path.remove(edge);
            }
        }
        return new HashMap<CapacityEdge, Integer>();
    }
 
    public int maxFlow(Node source, Node target) {
    	resetFlow();;
    	HashMap<CapacityEdge, Integer> path = findPath(source, target, new HashMap<CapacityEdge, Integer>());
        while (!path.isEmpty()) {
            int fl = Integer.MAX_VALUE;
            for (Edge edge : path.keySet()) {
            	int res = path.get(edge).intValue();
            	if (res < fl)
            		fl = res;
            }
            for (CapacityEdge edge : path.keySet()) {
                flow.put(edge, flow.get(edge).intValue() + fl);
                flow.put(edge.getRedge(), flow.get(edge.getRedge()).intValue() - fl);
            }
            path = findPath(source, target, new HashMap<CapacityEdge, Integer>());
        }
        int sum = 0;
        for (CapacityEdge edge : getEdges(source))
        	sum += flow.get(edge).intValue();
        return sum;
    }
    
    public CountHashMap findMatchingGateways(String gw_id) {
    	Node gw = findNode(gw_id);
		if (end_event != null && start_event!= null)
			addEdge("-1",end_event,start_event);
    	CountHashMap res = new CountHashMap();
    	for (Node gw2 : gateways) {
    		if (gw.equals(gw2))
    			continue;
    		if (maxFlow(gw,gw2) > 1 || maxFlow(gw2,gw) > 1)
    			res.put(gw2.getId(), 0);
    	}
    	removeEdge("-1");
    	return res;
    }
    public ArrayList<String> findMatchingBlock(String gw1_id, String gw2_id) {
    	Node gw1 = findNode(gw1_id);
    	Node gw2 = findNode(gw2_id);
    	ArrayList<String> result = new ArrayList<String>();
    	result.add(gw1_id);
    	boolean found = addChildren(result,gw1,gw2);
    	result.add(gw2_id);
    	if (found)
    		return result;
    	else
    		return new ArrayList<String>();
    }
    private boolean addChildren(ArrayList<String> list, Node node, Node end) {
    	boolean found = false;
    	for (int i=0; i<network.get(node).size(); i++) {
    		Edge edge = network.get(node).get(i);
			list.add(edge.getId());
			if (edge.getTarget().equals(end))
				found = true;
			else if (!list.contains(edge.getTarget().getId())) {
    			list.add(edge.getTarget().getId());
    			found = found || addChildren(list,edge.getTarget(),end);
    		}
    	}
    	return found;
    }
}
