package org.processmining.plugins.cheetah.statistics;

import java.util.*;
import java.util.Map.Entry;

import org.deckfour.xes.model.*;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class ProcessModel {

	public HashMap<String, Node> activities = new HashMap<String, Node>();
	public HashMap<String, Node> events = new HashMap<String, Node>();
	public HashMap<String, Node> xors = new HashMap<String, Node>();
	public HashMap<String, Node> ands = new HashMap<String, Node>();
	public HashMap<String, Edge> edges = new HashMap<String, Edge>();
	public HashMap<Node, ArrayList<Edge>> model = new HashMap<Node, ArrayList<Edge>>();
	
	HashMap<Node, Integer> connectors = null;
	
	public ProcessModel(XTrace trace) {
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event).toUpperCase();
			if (!event.getAttributes().containsKey("id"))
				continue;
			String id = event.getAttributes().get("id").toString();
			if (activity.equals("CREATE_ACTIVITY"))
				activities.put(id, new Node(id,"NODE"));
			else if (activity.equals("CREATE_START_EVENT"))
				events.put(id, new Node(id,"NODE"));
			else if (activity.equals("CREATE_END_EVENT"))
				events.put(id, new Node(id,"NODE"));
			else if (activity.equals("CREATE_XOR"))
				xors.put(id, new Node(id,"XOR"));
			else if (activity.equals("CREATE_AND"))
				ands.put(id, new Node(id,"AND"));
			else if (activity.equals("CREATE_EDGE")) {
				Node source = findNode(event.getAttributes().get("source_node_id").toString());
				Node target = findNode(event.getAttributes().get("target_node_id").toString());
				Edge edge = new Edge(id, source, target);
				edges.put(id, edge);
				if (!model.containsKey(source))
					model.put(source, new ArrayList<Edge>());
				model.get(source).add(edge);
			}
			else if (activity.equals("DELETE_ACTIVITY"))
				activities.remove(id);
			else if (activity.equals("DELETE_START_EVENT"))
				events.remove(id);
			else if (activity.equals("DELETE_END_EVENT"))
				events.remove(id);
			else if (activity.equals("DELETE_XOR"))
				xors.remove(id);
			else if (activity.equals("DELETE_AND"))
				ands.remove(id);
			else if (activity.equals("DELETE_EDGE")) {
				removeEdge(id);
				edges.remove(id);
			}
			else if (activity.equals("RECONNECT_EDGE")) {
				Node source = findNode(event.getAttributes().get("source_node_id").toString());
				Node target = findNode(event.getAttributes().get("target_node_id").toString());
				removeEdge(id);
				edges.remove(id);
				Edge edge = new Edge(id, source, target);
				edges.put(id, edge);
				if (!model.containsKey(source))
					model.put(source, new ArrayList<Edge>());
				model.get(source).add(edge);
			}
		}
    }
	
	private Node findNode(String id) {
		if (activities.get(id) != null)
			return activities.get(id);
		if (events.get(id) != null)
			return events.get(id);
		if (xors.get(id) != null)
			return xors.get(id);
		if (ands.get(id) != null)
			return ands.get(id);
		return null;
	}
	private void removeEdge(String id) {
		ArrayList<Edge> edgesToRemove = new ArrayList<Edge>();
		for (Node node : model.keySet()) {
			for (Edge edge: model.get(node)) {
				if (edge.getId().equals(id))
					edgesToRemove.add(edge);
			}
			for (Edge edge : edgesToRemove)
				model.get(node).remove(edge);
			edgesToRemove.clear();
		}
	}
	
    //#NODES
    public int getNumNodes() {
    	return activities.size() + events.size() + xors.size() + ands.size();
    }
    //#ARCS
    public int getNumArcs() {
    	return edges.size();
    }
    //#TASKS
    public int getNumTasks() {
    	return activities.size() + events.size();
    }
    //#CONNECTORS
    public int getNumConnectors() {
    	return xors.size() + ands.size();
    }
    //#AND-SPLITS
    public int getNumAndSplits() {
    	int num = 0;
    	for(Entry<String, Node> nodeSet : ands.entrySet()) {
    		Node and = nodeSet.getValue();
    		if (model.containsKey(and) && model.get(and).size() > 1)
    			num++;
    	}
    	return num;
    }
	//#AND-JOINS
    public int getNumAndJoins() {
    	setConnectors();
    	
    	int num = 0;
    	for(Entry<String, Node> nodeSet : ands.entrySet()) {
    		Node and = nodeSet.getValue();
    		if (model.containsKey(and) && (connectors.get(and).intValue() - model.get(and).size()) > 1)
    			num++;
    	}
    	return num;
    }
	//#XOR-SPLITS
    public int getNumXorSplits() {
    	int num = 0;
    	for(Entry<String, Node> nodeSet : xors.entrySet()) {
    		Node xor = nodeSet.getValue();
    		if (model.containsKey(xor) && model.get(xor).size() > 1)
    			num++;
    	}
    	return num;
    }
    //#XOR-JOINS
    public int getNumXorJoins() {
    	setConnectors();
    	
    	int num = 0;
    	for(Entry<String, Node> nodeSet : xors.entrySet()) {
    		Node xor = nodeSet.getValue();
    		if (model.containsKey(xor) && (connectors.get(xor).intValue() - model.get(xor).size()) > 1)
    			num++;
    	}
    	return num;
    }
    //DIAMETER (longest path from start to end without loops)
    public int getDiameter() {
    	ArrayList<Node> startNodes = getStartNodes();
    	int longestPath = 0;
    	for (Node node : startNodes) {
    		ArrayList<Node> path = getMaximumPath(node);
    		if (path.size() > longestPath)
    			longestPath = path.size();
    	}
    	return longestPath;
    }
    private ArrayList<Node> getStartNodes() {
    	ArrayList<Node> allNodes = new ArrayList<Node>();
    	for (Entry<String, Node> node : activities.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Node> node : events.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Node> node : xors.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Node> node : ands.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Edge> edge : edges.entrySet())
    		allNodes.remove(edge.getValue().getTarget());
    	return allNodes;
    }
    private ArrayList<Node> getMaximumPath(Node startNode) {
    	ArrayList<Node> maxPath = new ArrayList<Node>();
    	int longestPath = 0;
    	for (Edge edge : model.get(startNode)) {
    		ArrayList<Node> path = getMaximumPath(edge.getTarget());
    		if (!path.contains(startNode) && path.size() > longestPath) {
    			longestPath = path.size();
    			maxPath = path;
    		}
    	}
    	return maxPath;
    }
    //TOKENSPLITS
    public int getTokenSplits() {
    	int maxSplitPaths = 0;
    	for (Node node : model.keySet()) {
    		if (model.get(node).size() > maxSplitPaths)
    			maxSplitPaths = model.get(node).size(); 
    	}
    	return maxSplitPaths;
    }
    private void setConnectors() {
    	if (connectors != null)
    		return;
    	
    	connectors = new HashMap<Node, Integer>();
    	for (Entry<String, Edge> edgeSet : edges.entrySet()) {
    		Edge edge = edgeSet.getValue();
    		registerConnector(edge.getSource());
    		registerConnector(edge.getTarget());
    	}
    }
    private void registerConnector(Node node) {
    	if (!connectors.containsKey(node))
    		connectors.put(node, new Integer(1));
    	else
    		connectors.put(node, new Integer(connectors.get(node).intValue() + 1));
    }
    //AVG CONNECTOR DEGREE
    public double getAvgConnectorDegree() {
    	setConnectors();
    	
    	double total = 0.0;
    	for (Entry<Node, Integer> nodeSet : connectors.entrySet()) {
   			total += nodeSet.getValue().intValue();
   			System.out.println(nodeSet.getValue().intValue());
    	}
    	System.out.println("-----------\n" + 
    			           total +"/" + connectors.size() + "\n" + 
    			           (total / connectors.size()));
    	if (total / connectors.size() > 5)
    		System.out.println("GROOT?!");
    	return total / connectors.size();
    }
    //MAX CONNECTOR DEGREE
    public int getMaxConnectorDegree() {
    	setConnectors();
    	
    	int max = 0;
    	for (Entry<Node, Integer> nodeSet : connectors.entrySet()) {
    		if (nodeSet.getValue().intValue() > max)
    			max = nodeSet.getValue().intValue();
    	}
    	return max;
    }
    //CONTROL FLOW COMPLEXITY
    //MISMATCH
    //DEPTH
    //CONNECTIVITY
    public double getConnectivity() {
    	return getNumArcs() * 1.0 / getNumNodes();
    }
    //DENSITY
    public double getDensity() {
    	double maxPossibleNumArcs = ( getNumNodes() * (getNumNodes() - 1) ) / 2.0;
    	return getNumArcs() / maxPossibleNumArcs;
    }
    //CROSS-CONNECTIVITY
    //SEQUENTIALITY
    //SEPARABILITY
    public double getSeparability() {
    	return getNumCutVertices() / getNumNodes();
    }
    private int getNumCutVertices() {
    	ArrayList<Node> allNodes = new ArrayList<Node>();
    	for (Entry<String, Node> node : activities.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Node> node : events.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Node> node : xors.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Node> node : ands.entrySet())
    		allNodes.add(node.getValue());
    	for (Entry<String, Edge> edge : edges.entrySet())
    		allNodes.remove(edge.getValue().getTarget());
    	int getNumCutVertices = 0;
    	
    	for (Node node : allNodes) {
    		ArrayList<Node> inNodes = getInNodes(node); inNodes.remove(node);
    		ArrayList<Node> outNodes = getOutNodes(node); outNodes.remove(node);
    		
    		if (getIntersection(inNodes, outNodes).size() == 0)
    			getNumCutVertices++;
    	}
    	return getNumCutVertices;
    }
    private ArrayList<Node> getInNodes(Node node) {
    	ArrayList<Node> inNodes = new ArrayList<Node>();
    	inNodes.add(node);
    	for (Entry<String, Edge> edge : edges.entrySet()) {
    		if (!edge.getValue().getTarget().equals(node))
    			continue;
    		ArrayList<Node> nodes = getInNodes(edge.getValue().getSource());
    		for(Node n : nodes)
    			if (!inNodes.contains(n))
    				inNodes .add(n);
    	}
    	return inNodes;
    }
    private ArrayList<Node> getOutNodes(Node node) {
    	ArrayList<Node> outNodes = new ArrayList<Node>();
    	outNodes.add(node);
    	for (Edge edge : model.get(node)) {
    		ArrayList<Node> nodes = getOutNodes(edge.getTarget());
    		for(Node n : nodes)
    			if (!outNodes.contains(n))
    				outNodes .add(n);
    	}
    	return outNodes;
    }
    private ArrayList<Node> getIntersection(ArrayList<Node> list1, ArrayList<Node> list2) {
    	ArrayList<Node> intersection = new ArrayList<Node>();
    	for(Node node : list1)
    		if (list2.contains(node))
    			intersection.add(node);
    	return intersection;
    }
    //STRUCTUREDNESS
    //CONNECTOR HETEROGENEITY
    
}
